"""
TimeSeries Datasets
-------------------
"""

try:
    # Base classes for training datasets:
    # Implementation (horizon-based)
    from darts.utils.data.horizon_based_dataset import HorizonBasedDataset

    # Base class and implementations for inference datasets:
    from darts.utils.data.inference_dataset import (
        DualCovariatesInferenceDataset,
        FutureCovariatesInferenceDataset,
        InferenceDataset,
        MixedCovariatesInferenceDataset,
        PastCovariatesInferenceDataset,
        SplitCovariatesInferenceDataset,
    )

    # Implementations (sequential)
    from darts.utils.data.sequential_dataset import (
        DualCovariatesSequentialDataset,
        FutureCovariatesSequentialDataset,
        MixedCovariatesSequentialDataset,
        PastCovariatesSequentialDataset,
        SplitCovariatesSequentialDataset,
    )

    # Implementations (shifted)
    from darts.utils.data.shifted_dataset import (
        DualCovariatesShiftedDataset,
        FutureCovariatesShiftedDataset,
        MixedCovariatesShiftedDataset,
        PastCovariatesShiftedDataset,
        SplitCovariatesShiftedDataset,
    )
    from darts.utils.data.training_dataset import (
        DualCovariatesTrainingDataset,
        FutureCovariatesTrainingDataset,
        MixedCovariatesTrainingDataset,
        PastCovariatesTrainingDataset,
        SplitCovariatesTrainingDataset,
        TrainingDataset,
    )
except ImportError:  # Torch is not available
    from darts.utils.utils import NotImportedModule

    HorizonBasedDataset = NotImportedModule(module_name="(Py)Torch", warn=False)
    DualCovariatesInferenceDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    FutureCovariatesInferenceDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    InferenceDataset = NotImportedModule(module_name="(Py)Torch", warn=False)
    MixedCovariatesInferenceDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    PastCovariatesInferenceDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    SplitCovariatesInferenceDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    DualCovariatesSequentialDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    FutureCovariatesSequentialDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    MixedCovariatesSequentialDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    PastCovariatesSequentialDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    SplitCovariatesSequentialDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    DualCovariatesShiftedDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    FutureCovariatesShiftedDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    MixedCovariatesShiftedDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    PastCovariatesShiftedDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    SplitCovariatesShiftedDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    DualCovariatesTrainingDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    FutureCovariatesTrainingDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    MixedCovariatesTrainingDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    PastCovariatesTrainingDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    SplitCovariatesTrainingDataset = NotImportedModule(
        module_name="(Py)Torch", warn=False
    )
    TrainingDataset = NotImportedModule(module_name="(Py)Torch", warn=False)

__all__ = [
    "HorizonBasedDataset",
    "DualCovariatesInferenceDataset",
    "FutureCovariatesInferenceDataset",
    "InferenceDataset",
    "MixedCovariatesInferenceDataset",
    "PastCovariatesInferenceDataset",
    "SplitCovariatesInferenceDataset",
    "DualCovariatesSequentialDataset",
    "FutureCovariatesSequentialDataset",
    "MixedCovariatesSequentialDataset",
    "PastCovariatesSequentialDataset",
    "SplitCovariatesSequentialDataset",
    "DualCovariatesShiftedDataset",
    "FutureCovariatesShiftedDataset",
    "MixedCovariatesShiftedDataset",
    "PastCovariatesShiftedDataset",
    "SplitCovariatesShiftedDataset",
    "DualCovariatesTrainingDataset",
    "FutureCovariatesTrainingDataset",
    "MixedCovariatesTrainingDataset",
    "PastCovariatesTrainingDataset",
    "SplitCovariatesTrainingDataset",
    "TrainingDataset",
]



"""
StatsForecastModel
------------------
"""

from typing import Optional

import numpy as np
from statsforecast.models import _TS

from darts import TimeSeries, concatenate
from darts.logging import get_logger
from darts.models import LinearRegressionModel
from darts.models.forecasting.forecasting_model import (
    TransferableFutureCovariatesLocalForecastingModel,
)
from darts.utils.likelihood_models.statsforecast import QuantilePrediction
from darts.utils.timeseries_generation import _build_forecast_series
from darts.utils.utils import random_method

logger = get_logger(__name__)


class StatsForecastModel(TransferableFutureCovariatesLocalForecastingModel):
    @random_method
    def __init__(
        self,
        model: _TS,
        add_encoders: Optional[dict] = None,
        quantiles: Optional[list[float]] = None,
        random_state: Optional[int] = None,
    ):
        """StatsForecast Model.

        Can be used to fit any `StatsForecast` base model. For more information on available models, see the
        `StatsForecast package <https://nixtlaverse.nixtla.io/statsforecast/index.html>`_.

        In addition to univariate deterministic forecasting, our `StatsForecastModel` comes with additional support:

        - **Future covariates:** Use exogenous features to potentially improve predictive accuracy.

          - It either uses the base model's native exogenous features, or

          - It adds future covariates support by first regressing the series against the future covariates using a
            :class:`~darts.models.forecasting.linear_regression_model.LinearRegressionModel` model and then running the
            StatsForecast model on the in-sample residuals from this original regression. This approach was inspired by
            `this post of Stephan Kolassa <https://stats.stackexchange.com/q/220885>`_.

        - **Probabilstic forecasting:** Some base models might require setting `prediction_intervals` at `model`
          creation to support probabilistic forecasting. To generate probabilistic forecasts, you can set the following
          parameters when calling :meth:`~darts.models.forecasting.sf_model.StatsForecastModel.predict`:

          - Forecast quantile values directly by setting `predict_likelihood_parameters=True`.

          - Generate sampled forecasts from these quantiles by setting `num_samples >> 1`.

        - **Conformal prediction:** In addition to the native probabilistic support, you can perform conformal
          prediction / forecasting by setting `prediction_intervals` at model creation. Then predict the in the same
          way as described above.

        - **Transferable series forecasting:** Apply the fitted model to a new input `series` at prediction time.

          - It either uses the base model's native transferrable series support (StatsForecast models that support the
            `forward()` method), or

          - It adds support by first fitting a copy of the model on the new series, and then using that model to
            generate the corresponding forecast.

        .. note::
            Future covariates are not supported when the input series contain missing values.

        Parameters
        ----------
        model
            Any StatsForecast model.
        add_encoders
            A large number of future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                def encode_year(idx):
                    return (idx.year - 1950) / 50

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'future': ['relative']},
                    'custom': {'future': [encode_year]},
                    'transformer': Scaler(),
                    'tz': 'CET'
                }
            ..
        quantiles
            Optionally, produce quantile predictions at `quantiles` levels when performing probabilistic forecasting
            with `num_samples > 1` or `predict_likelihood_parameters=True`.
        random_state
            Control the randomness of probabilistic conformal forecasts (sample generation) across different runs.

        Examples
        --------
        >>> from darts.datasets import AirPassengersDataset
        >>> from darts.models import StatsForecastModel
        >>> from darts.utils.timeseries_generation import datetime_attribute_timeseries
        >>> from statsforecast.models import AutoARIMA
        >>> series = AirPassengersDataset().load()
        >>> # optionally, use some future covariates; e.g. the value of the month encoded as a sine and cosine series
        >>> future_cov = datetime_attribute_timeseries(series, "month", cyclic=True, add_length=6)
        >>> # define AutoARIMA parameters
        >>> model = StatsForecastModel(model=AutoARIMA(season_length=12))
        >>> model.fit(series, future_covariates=future_cov)
        >>> pred = model.predict(6, future_covariates=future_cov)
        >>> pred.values()
        array([[445.4276575 ],
               [420.04912881],
               [448.7142377 ],
               [491.23406559],
               [502.67834069],
               [566.04774778]])
        """
        if not isinstance(model, _TS):
            raise ValueError(
                "`model` must be a StatsForecast model imported from `statsforecast.models`."
            )
        self.model: _SFModel = model
        self._likelihood = QuantilePrediction(quantiles=quantiles)

        # future covariates support can be added through the use of a linear model
        self._linreg: Optional[LinearRegressionModel] = None
        super().__init__(add_encoders=add_encoders)

    def _fit(self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None):
        super()._fit(series, future_covariates)
        self._assert_univariate(series)
        series = self.training_series

        if self._supports_native_future_covariates or future_covariates is None:
            target = series.values(copy=False).flatten()
        else:
            # perform OLS and get in-sample residuals
            self._linreg = LinearRegressionModel(lags_future_covariates=[0])
            self._linreg.fit(series, future_covariates=future_covariates)
            target = self._get_target_residuals(series, future_covariates)

        self.model.fit(
            target,
            X=(
                future_covariates.values(copy=False)
                if future_covariates is not None
                and self._supports_native_future_covariates
                else None
            ),
        )
        return self

    @random_method
    def _predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        historic_future_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        predict_likelihood_parameters: bool = False,
        verbose: bool = False,
    ) -> TimeSeries:
        super()._predict(
            n, series, historic_future_covariates, future_covariates, num_samples
        )

        if series is not None and not self._supports_native_transferable_series:
            # if model doesn't support transferable series forecasts, we fit and predict using a copy of the model
            if future_covariates is not None:
                # merge covariates
                future_covariates = concatenate(
                    [historic_future_covariates, future_covariates], axis=0
                )

                encoders = self.encoders
                if encoders is not None and encoders.encoding_available:
                    # drop encoded covariates from the covariates
                    future_covariates = encoders._drop_encoded_components(
                        covariates=future_covariates,
                        components=encoders.future_components,
                    )

            return (
                self.untrained_model()
                .fit(
                    series=series,
                    future_covariates=future_covariates,
                )
                .predict(
                    n=n,
                    num_samples=num_samples,
                    predict_likelihood_parameters=predict_likelihood_parameters,
                    verbose=verbose,
                )
            )

        # confidence levels for prediction intervals
        levels = (
            self.likelihood.levels
            if num_samples > 1 or predict_likelihood_parameters
            else None
        )
        model_output = self._estimator_predict(
            n=n,
            series=series,
            historic_future_covariates=historic_future_covariates,
            future_covariates=future_covariates,
            levels=levels,
        )

        series = series if series is not None else self.training_series

        # likelihood takes care of probabilistic forecasts
        pred_vals = self.likelihood.predict(model_output, num_samples=num_samples)

        comp_names_out = None
        if predict_likelihood_parameters:
            comp_names_out = self.likelihood.component_names(series)

        return _build_forecast_series(
            points_preds=pred_vals,
            input_series=series,
            custom_columns=comp_names_out,
            with_static_covs=not predict_likelihood_parameters,
            with_hierarchy=not predict_likelihood_parameters,
        )

    def _estimator_predict(
        self,
        n: int,
        series: Optional[TimeSeries],
        historic_future_covariates: Optional[TimeSeries],
        future_covariates: Optional[TimeSeries],
        levels: Optional[list[float]],
    ) -> np.ndarray:
        """
        Computes the model output.

        When this method is called, it is guaranteed that either `series` is None, or that the
        model supports transferable series forecasting (model has a forward method).

        Parameters
        ----------
        n
            The number of time steps after the end of the training time series for which to produce predictions
        series
            The series whose future values will be predicted. If the statsforecast model has a `forward` method then
            the previously fitted model will be used for prediction, otherwise a copy of the model will be fitted to
            predict the new series.
        historic_future_covariates
            Optionally, the historic part of the future-known covariates series that is overlapping with `series`.
            They must match the covariates used for training in terms of dimension.
        future_covariates
            Optionally, the future-known covariates series that extends to at least `n` steps after the end of `series`.
            They must match the covariates used for training in terms of dimension.
        levels
            The confidence levels (0. - 100.) for the prediction intervals.
        """
        native_cov_support = (
            future_covariates is not None and self._supports_native_future_covariates
        )
        custom_cov_support = (
            future_covariates is not None
            and not self._supports_native_future_covariates
        )

        x_future = future_covariates.values(copy=False) if native_cov_support else None

        if series is None:
            forecast_dict = self.model.predict(h=n, X=x_future, level=levels)
        else:
            # model has a `forward` method, it supports transferable prediction series (uses
            # fitted model to forecast new series)
            if custom_cov_support:
                target = self._get_target_residuals(
                    series,
                    historic_future_covariates,
                )
            else:
                target = series.values(copy=False).flatten()

            x_historic = (
                historic_future_covariates.values(copy=False)
                if native_cov_support
                else None
            )
            forecast_dict = self.model.forward(
                y=target, h=n, X=x_historic, X_future=x_future, level=levels
            )

        vals = _unpack_sf_dict(forecast_dict, levels=levels)
        if custom_cov_support:
            # sf model was trained on residuals, add back the remainder
            mu_linreg = self._linreg.predict(
                n, series=series, future_covariates=future_covariates
            )
            mu_linreg_values = mu_linreg.values(copy=False).reshape(n, 1)
            vals += mu_linreg_values
        return vals

    def _get_target_residuals(
        self,
        series: TimeSeries,
        future_covariates: TimeSeries,
    ) -> np.ndarray:
        """Computes the OLS residuals for predicting the target series from `future_covariates`."""
        fitted_values = self._linreg.model.predict(
            X=future_covariates.slice_intersect_values(series, copy=False)[:, :, 0]
        )
        residuals = series.values(copy=False).flatten() - fitted_values
        return residuals

    @property
    def supports_multivariate(self) -> bool:
        return False

    @property
    def min_train_series_length(self) -> int:
        return 10

    @property
    def _supports_range_index(self) -> bool:
        return True

    @property
    def supports_probabilistic_prediction(self) -> bool:
        return self.likelihood is not None

    @property
    def _supports_native_transferable_series(self) -> bool:
        return hasattr(self.model, "forward")

    @property
    def _supports_native_future_covariates(self) -> bool:
        return self.model.uses_exog

    @property
    def likelihood(self) -> QuantilePrediction:
        return self._likelihood

    @property
    def _supports_non_retrainable_historical_forecasts(self) -> bool:
        return self._supports_native_transferable_series


class _SFModel(_TS):
    """This serves as a protocol for expected StatsForecast model API."""

    def fit(self, *args, **kwargs): ...

    def predict(self, *args, **kwargs) -> dict: ...

    def forecast(self, *args, **kwargs) -> dict: ...

    def forward(self, *args, **kwargs) -> dict: ...


def _unpack_sf_dict(
    forecast_dict: dict,
    levels: Optional[list[float]],
) -> np.ndarray:
    """Unpack the dictionary that is returned by the StatsForecast 'predict()' method.

    Into an array of quantile predictions with shape (n (horizon), n quantiles) ordered by increasing quantile.
    """
    mu = np.expand_dims(forecast_dict["mean"], -1)
    if levels is None:
        return mu

    lows = np.concatenate(
        [np.expand_dims(forecast_dict[f"lo-{level}"], -1) for level in levels], axis=1
    )
    highs = np.concatenate(
        [np.expand_dims(forecast_dict[f"hi-{level}"], -1) for level in levels[::-1]],
        axis=1,
    )
    return np.concatenate([lows, mu, highs], axis=1)



"""
Recurrent Neural Networks
-------------------------
"""

import inspect
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Optional, Union

import torch
import torch.nn as nn

from darts.logging import get_logger, raise_if_not, raise_log
from darts.models.forecasting.pl_forecasting_module import (
    PLDualCovariatesModule,
    io_processor,
)
from darts.models.forecasting.torch_forecasting_model import DualCovariatesTorchModel
from darts.timeseries import TimeSeries
from darts.utils.data import DualCovariatesShiftedDataset, TrainingDataset

logger = get_logger(__name__)


class CustomRNNModule(PLDualCovariatesModule, ABC):
    def __init__(
        self,
        input_size: int,
        hidden_dim: int,
        num_layers: int,
        target_size: int,
        nr_params: int,
        dropout: float = 0.0,
        **kwargs,
    ):
        """This class allows to create custom RNN modules that can later be used with Darts' :class:`RNNModel`.
        It adds the backbone that is required to be used with Darts' :class:`TorchForecastingModel` and
        :class:`RNNModel`.

        To create a new module, subclass from :class:`CustomRNNModule` and:

        * Define the architecture in the module constructor (`__init__()`)

        * Add the `forward()` method and define the logic of your module's forward pass

        * Use the custom module class when creating a new :class:`RNNModel` with parameter `model`.

        You can use `darts.models.forecasting.rnn_model._RNNModule` as an example.

        Parameters
        ----------
        input_size
            The dimensionality of the input time series.
        hidden_dim
            The number of features in the hidden state `h` of the RNN module.
        num_layers
            The number of recurrent layers.
        target_size
            The dimensionality of the output time series.
        nr_params
            The number of parameters of the likelihood (or 1 if no likelihood is used).
        dropout
            The fraction of neurons that are dropped in all-but-last RNN layers.
        **kwargs
            all parameters required for :class:`darts.models.forecasting.pl_forecasting_module.PLForecastingModule`
            base class.
        """
        # RNNModule doesn't really need input and output_chunk_length for PLModule
        super().__init__(**kwargs)

        # Defining parameters
        self.input_size = input_size
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.target_size = target_size
        self.nr_params = nr_params
        self.dropout = dropout

    @io_processor
    @abstractmethod
    def forward(
        self, x_in: tuple, h: Optional[torch.Tensor] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """RNN Module forward.

        Parameters
        ----------
        x_in
            Tuple of Tensors containing the features of the input sequence. The tuple has elements (past target,
            historic future covariates, future covariates, static covariates). The shape of the past target is
            `(batch_size, input_length, input_size)`.
        h
            Optionally, the hidden state.

        Returns
        -------
        Tuple[torch.Tensor, torch.Tensor]
            Tuple of Tensors with elements (RNN output, hidden state). The RNN output Tensor has shape
            `(batch_size, output_chunk_length, target_size, nr_params)`. It contains the outputs at every
            time step of the input sequence. During training the whole tensor is used as output, whereas during
            prediction we only use y[:, -1, :]. However, this module always returns the whole Tensor.
        """
        pass

    def _produce_train_output(self, input_batch: tuple) -> torch.Tensor:
        # only return the forecast, not the hidden state
        return self(self._process_input_batch(input_batch))[0]

    def _process_input_batch(
        self, input_batch: tuple
    ) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
        (
            past_target,
            historic_future_covariates,
            future_covariates,
            static_covariates,
        ) = input_batch
        # For the RNN we concatenate the past_target with the future_covariates
        # (they have the same length because we enforce a Shift dataset for RNNs)
        return (
            (
                torch.cat([past_target, future_covariates], dim=2)
                if future_covariates is not None
                else past_target
            ),
            static_covariates,
        )

    def _produce_predict_output(
        self, x: tuple, last_hidden_state: Optional[torch.Tensor] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """overwrite parent classes `_produce_predict_output` method"""
        output, hidden = self(x, last_hidden_state)
        if self.likelihood:
            if self.predict_likelihood_parameters:
                return self.likelihood.predict_likelihood_parameters(output), hidden
            else:
                return self.likelihood.sample(output), hidden
        else:
            return output.squeeze(dim=-1), hidden

    def _get_batch_prediction(
        self, n: int, input_batch: tuple, roll_size: int
    ) -> torch.Tensor:
        """
        This model is recurrent, so we have to write a specific way to
        obtain the time series forecasts of length n.
        """
        (
            past_target,
            historic_future_covariates,
            future_covariates,
            static_covariates,
        ) = input_batch

        if historic_future_covariates is not None:
            # RNNs need as inputs (target[t] and covariates[t+1]) so here we shift the covariates
            all_covariates = torch.cat(
                [historic_future_covariates[:, 1:, :], future_covariates], dim=1
            )
            cov_past, cov_future = (
                all_covariates[:, : past_target.shape[1], :],
                all_covariates[:, past_target.shape[1] :, :],
            )
            input_series = torch.cat([past_target, cov_past], dim=2)
        else:
            input_series = past_target
            cov_future = None

        batch_prediction = []
        out, last_hidden_state = self._produce_predict_output((
            input_series,
            static_covariates,
        ))
        batch_prediction.append(out[:, -1:, :])
        prediction_length = 1

        while prediction_length < n:
            # create new input to model from last prediction and current covariates, if available
            new_input = (
                torch.cat(
                    [
                        out[:, -1:, :],
                        cov_future[:, prediction_length - 1 : prediction_length, :],
                    ],
                    dim=2,
                )
                if cov_future is not None
                else out[:, -1:, :]
            )

            # feed new input to model, including the last hidden state from the previous iteration
            out, last_hidden_state = self._produce_predict_output(
                (new_input, static_covariates), last_hidden_state
            )

            # append prediction to batch prediction array, increase counter
            batch_prediction.append(out[:, -1:, :])
            prediction_length += 1

        # bring predictions into desired format and drop unnecessary values
        batch_prediction = torch.cat(batch_prediction, dim=1)
        batch_prediction = batch_prediction[:, :n, :]
        return batch_prediction


# TODO add batch norm
class _RNNModule(CustomRNNModule):
    def __init__(
        self,
        name: str,
        **kwargs,
    ):
        """PyTorch module implementing an RNN to be used in `RNNModel`.

        PyTorch module implementing a simple RNN with the specified `name` type.
        This module combines a PyTorch RNN module, together with one fully connected layer which
        maps the hidden state of the RNN at each step to the output value of the model at that
        time step.

        Parameters
        ----------
        name
            The name of the specific PyTorch RNN module ("RNN", "GRU" or "LSTM").
        **kwargs
            all parameters required for the :class:`darts.models.forecasting.CustomRNNModule` base class.

        Inputs
        ------
        x of shape `(batch_size, input_length, input_size)`
            Tensor containing the features of the input sequence. The `input_length` is not fixed.

        Outputs
        -------
        y of shape `(batch_size, output_chunk_length, target_size, nr_params)`
            Tensor containing the outputs of the RNN at every time step of the input sequence.
            During training the whole tensor is used as output, whereas during prediction we only use y[:, -1, :].
            However, this module always returns the whole Tensor.
        """

        # RNNModule doesn't really need input and output_chunk_length for PLModule
        super().__init__(**kwargs)
        self.name = name

        # Defining the RNN module
        self.rnn = getattr(nn, name)(
            self.input_size,
            self.hidden_dim,
            self.num_layers,
            batch_first=True,
            dropout=self.dropout,
        )

        # The RNN module needs a linear layer V that transforms hidden states into outputs, individually
        self.V = nn.Linear(self.hidden_dim, self.target_size * self.nr_params)

    @io_processor
    def forward(
        self, x_in: tuple, h: Optional[torch.Tensor] = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        x, _ = x_in
        # data is of size (batch_size, input_length, input_size)
        batch_size = x.shape[0]

        # out is of size (batch_size, input_length, hidden_dim)
        out, last_hidden_state = self.rnn(x) if h is None else self.rnn(x, h)

        # Here, we apply the V matrix to every hidden state to produce the outputs
        predictions = self.V(out)

        # predictions is of size (batch_size, input_length, target_size)
        predictions = predictions.view(batch_size, -1, self.target_size, self.nr_params)

        # returns outputs for all inputs, only the last one is needed for prediction time
        return predictions, last_hidden_state


class RNNModel(DualCovariatesTorchModel):
    def __init__(
        self,
        input_chunk_length: int,
        model: Union[str, type[CustomRNNModule]] = "RNN",
        hidden_dim: int = 25,
        n_rnn_layers: int = 1,
        dropout: float = 0.0,
        training_length: int = 24,
        **kwargs,
    ):
        """Recurrent Neural Network Model (RNNs).

        This class provides three variants of RNNs:

        * Vanilla RNN

        * LSTM

        * GRU

        RNNModel is fully recurrent in the sense that, at prediction time, an output is computed using these inputs:

        - previous target value, which will be set to the last known target value for the first prediction,
          and for all other predictions it will be set to the previous prediction (in an autoregressive fashion),
        - the previous hidden state,
        - the covariates at time `t` for forecasting the target at time `t` (if the model was trained with covariates),

        This model supports future covariates; and it requires these covariates to extend far enough in the past
        and the future (it's a so-called "dual covariates" model as the future covariates have to be provided both
        in the past and the future). The model will complain if the provided `future_covariates` series doesn't have
        an appropriate time span.

        For a block version using an RNN model as an encoder only and supporting past
        covariates, checkout `BlockRNNModel`.

        Parameters
        ----------
        input_chunk_length
            Number of past time steps that are fed to the forecasting module at prediction time.
        model
            Either a string specifying the RNN module type ("RNN", "LSTM" or "GRU"), or a subclass of
            :class:`CustomRNNModule` (the class itself, not an object of the class) with a custom logic.
        hidden_dim
            Size for feature maps for each hidden RNN layer (:math:`h_n`).
        n_rnn_layers
            The number of recurrent layers.
        dropout
            Fraction of neurons affected by Dropout.
        training_length
            The length of both input (target and covariates) and output (target) time series used during
            training. Must have a larger value than `input_chunk_length`, because otherwise during training
            the RNN is never run for as many iterations as it will during inference. For more information on
            this parameter, please see `darts.utils.data.ShiftedDataset`.
        **kwargs
            Optional arguments to initialize the pytorch_lightning.Module, pytorch_lightning.Trainer, and
            Darts' :class:`TorchForecastingModel`.

        loss_fn
            PyTorch loss function used for training.
            This parameter will be ignored for probabilistic models if the ``likelihood`` parameter is specified.
            Default: ``torch.nn.MSELoss()``.
        likelihood
            One of Darts' :meth:`Likelihood <darts.utils.likelihood_models.torch.TorchLikelihood>` models to be used for
            probabilistic forecasts. Default: ``None``.
        torch_metrics
            A torch metric or a ``MetricCollection`` used for evaluation. A full list of available metrics can be found
            at https://torchmetrics.readthedocs.io/en/latest/. Default: ``None``.
        optimizer_cls
            The PyTorch optimizer class to be used. Default: ``torch.optim.Adam``.
        optimizer_kwargs
            Optionally, some keyword arguments for the PyTorch optimizer (e.g., ``{'lr': 1e-3}``
            for specifying a learning rate). Otherwise the default values of the selected ``optimizer_cls``
            will be used. Default: ``None``.
        lr_scheduler_cls
            Optionally, the PyTorch learning rate scheduler class to be used. Specifying ``None`` corresponds
            to using a constant learning rate. Default: ``None``.
        lr_scheduler_kwargs
            Optionally, some keyword arguments for the PyTorch learning rate scheduler. Default: ``None``.
        batch_size
            Number of time series (input and output sequences) used in each training pass. Default: ``32``.
        n_epochs
            Number of epochs over which to train the model. Default: ``100``.
        model_name
            Name of the model. Used for creating checkpoints and saving tensorboard data. If not specified,
            defaults to the following string ``"YYYY-mm-dd_HH_MM_SS_torch_model_run_PID"``, where the initial part
            of the name is formatted with the local date and time, while PID is the processed ID (preventing models
            spawned at the same time by different processes to share the same model_name). E.g.,
            ``"2021-06-14_09_53_32_torch_model_run_44607"``.
        work_dir
            Path of the working directory, where to save checkpoints and Tensorboard summaries.
            Default: current working directory.
        log_tensorboard
            If set, use Tensorboard to log the different parameters. The logs will be located in:
            ``"{work_dir}/darts_logs/{model_name}/logs/"``. Default: ``False``.
        nr_epochs_val_period
            Number of epochs to wait before evaluating the validation loss (if a validation
            ``TimeSeries`` is passed to the :func:`fit()` method). Default: ``1``.
        force_reset
            If set to ``True``, any previously-existing model with the same name will be reset (all checkpoints will
            be discarded). Default: ``False``.
        save_checkpoints
            Whether to automatically save the untrained model and checkpoints from training.
            To load the model from checkpoint, call :func:`MyModelClass.load_from_checkpoint()`, where
            :class:`MyModelClass` is the :class:`TorchForecastingModel` class that was used (such as :class:`TFTModel`,
            :class:`NBEATSModel`, etc.). If set to ``False``, the model can still be manually saved using
            :func:`save()` and loaded using :func:`load()`. Default: ``False``.
        add_encoders
            A large number of past and future covariates can be automatically generated with `add_encoders`.
            This can be done by adding multiple pre-defined index encoders and/or custom user-made functions that
            will be used as index encoders. Additionally, a transformer such as Darts' :class:`Scaler` can be added to
            transform the generated covariates. This happens all under one hood and only needs to be specified at
            model creation.
            Read :meth:`SequentialEncoder <darts.dataprocessing.encoders.SequentialEncoder>` to find out more about
            ``add_encoders``. Default: ``None``. An example showing some of ``add_encoders`` features:

            .. highlight:: python
            .. code-block:: python

                def encode_year(idx):
                    return (idx.year - 1950) / 50

                add_encoders={
                    'cyclic': {'future': ['month']},
                    'datetime_attribute': {'future': ['hour', 'dayofweek']},
                    'position': {'past': ['relative'], 'future': ['relative']},
                    'custom': {'past': [encode_year]},
                    'transformer': Scaler(),
                    'tz': 'CET'
                }
            ..
        random_state
            Control the randomness of the weights initialization. Check this
            `link <https://scikit-learn.org/stable/glossary.html#term-random_state>`_ for more details.
            Default: ``None``.
        pl_trainer_kwargs
            By default :class:`TorchForecastingModel` creates a PyTorch Lightning Trainer with several useful presets
            that performs the training, validation and prediction processes. These presets include automatic
            checkpointing, tensorboard logging, setting the torch device and more.
            With ``pl_trainer_kwargs`` you can add additional kwargs to instantiate the PyTorch Lightning trainer
            object. Check the `PL Trainer documentation
            <https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html>`_ for more information about the
            supported kwargs. Default: ``None``.
            Running on GPU(s) is also possible using ``pl_trainer_kwargs`` by specifying keys ``"accelerator",
            "devices", and "auto_select_gpus"``. Some examples for setting the devices inside the ``pl_trainer_kwargs``
            dict:

            - ``{"accelerator": "cpu"}`` for CPU,
            - ``{"accelerator": "gpu", "devices": [i]}`` to use only GPU ``i`` (``i`` must be an integer),
            - ``{"accelerator": "gpu", "devices": -1, "auto_select_gpus": True}`` to use all available GPUS.

            For more info, see here:
            https://pytorch-lightning.readthedocs.io/en/stable/common/trainer.html#trainer-flags , and
            https://pytorch-lightning.readthedocs.io/en/stable/accelerators/gpu_basic.html#train-on-multiple-gpus

            With parameter ``"callbacks"`` you can add custom or PyTorch-Lightning built-in callbacks to Darts'
            :class:`TorchForecastingModel`. Below is an example for adding EarlyStopping to the training process.
            The model will stop training early if the validation loss `val_loss` does not improve beyond
            specifications. For more information on callbacks, visit:
            `PyTorch Lightning Callbacks
            <https://pytorch-lightning.readthedocs.io/en/stable/extensions/callbacks.html>`_

            .. highlight:: python
            .. code-block:: python

                from pytorch_lightning.callbacks.early_stopping import EarlyStopping

                # stop training when validation loss does not decrease more than 0.05 (`min_delta`) over
                # a period of 5 epochs (`patience`)
                my_stopper = EarlyStopping(
                    monitor="val_loss",
                    patience=5,
                    min_delta=0.05,
                    mode='min',
                )

                pl_trainer_kwargs={"callbacks": [my_stopper]}
            ..

            Note that you can also use a custom PyTorch Lightning Trainer for training and prediction with optional
            parameter ``trainer`` in :func:`fit()` and :func:`predict()`.
        show_warnings
            whether to show warnings raised from PyTorch Lightning. Useful to detect potential issues of
            your forecasting use case. Default: ``False``.

        Examples
        --------
        >>> from darts.datasets import WeatherDataset
        >>> from darts.models import RNNModel
        >>> series = WeatherDataset().load()
        >>> # predicting atmospheric pressure
        >>> target = series['p (mbar)'][:100]
        >>> # optionally, use future temperatures (pretending this component is a forecast)
        >>> future_cov = series['T (degC)'][:106]
        >>> # `training_length` > `input_chunk_length` to mimic inference constraints
        >>> model = RNNModel(
        >>>     model="RNN",
        >>>     input_chunk_length=6,
        >>>     training_length=18,
        >>>     n_epochs=20,
        >>> )
        >>> model.fit(target, future_covariates=future_cov)
        >>> pred = model.predict(6)
        >>> pred.values()
        array([[ 3.18922903],
               [ 1.17791019],
               [ 0.39992814],
               [ 0.13277921],
               [ 0.02523252],
               [-0.01829086]])

        .. note::
            `RNN example notebook <https://unit8co.github.io/darts/examples/04-RNN-examples.html>`_ presents techniques
            that can be used to improve the forecasts quality compared to this simple usage example.
        """
        if training_length < input_chunk_length:
            raise_log(
                ValueError(
                    f"`training_length` ({training_length}) must be `>=input_chunk_length` ({input_chunk_length})."
                ),
                logger=logger,
            )
        # create copy of model parameters
        model_kwargs = {key: val for key, val in self.model_params.items()}

        for kwarg, default_value in zip(
            [
                "output_chunk_length",
                "use_reversible_instance_norm",
                "output_chunk_shift",
            ],
            [1, False, 0],
        ):
            if model_kwargs.get(kwarg) is not None:
                logger.warning(
                    f"ignoring user defined `{kwarg}`. RNNModel uses a fixed "
                    f"`{kwarg}={default_value}`."
                )
            model_kwargs[kwarg] = default_value

        super().__init__(**self._extract_torch_model_params(**model_kwargs))

        # extract pytorch lightning module kwargs
        self.pl_module_params = self._extract_pl_module_params(**model_kwargs)

        # check we got right model type specified:
        if model not in ["RNN", "LSTM", "GRU"]:
            if not inspect.isclass(model) or not issubclass(model, CustomRNNModule):
                raise_log(
                    ValueError(
                        "`model` is not a valid RNN model. Please specify 'RNN', 'LSTM', 'GRU', or give a subclass "
                        "(not an instance) of darts.models.forecasting.rnn_model.CustomRNNModule."
                    ),
                    logger=logger,
                )

        self.rnn_type_or_module = model
        self.dropout = dropout
        self.hidden_dim = hidden_dim
        self.n_rnn_layers = n_rnn_layers
        self.training_length = training_length

    def _create_model(self, train_sample: tuple[torch.Tensor]) -> torch.nn.Module:
        # samples are made of (past_target, historic_future_covariates, future_covariates, future_target)
        # historic_future_covariates and future_covariates have the same width
        input_dim = train_sample[0].shape[1] + (
            train_sample[1].shape[1] if train_sample[1] is not None else 0
        )
        output_dim = train_sample[-1].shape[1]
        nr_params = 1 if self.likelihood is None else self.likelihood.num_parameters

        kwargs = {}
        if isinstance(self.rnn_type_or_module, str):
            model_cls = _RNNModule
            kwargs["name"] = self.rnn_type_or_module
        else:
            model_cls = self.rnn_type_or_module
        return model_cls(
            input_size=input_dim,
            target_size=output_dim,
            nr_params=nr_params,
            hidden_dim=self.hidden_dim,
            dropout=self.dropout,
            num_layers=self.n_rnn_layers,
            **self.pl_module_params,
            **kwargs,
        )

    def _build_train_dataset(
        self,
        target: Sequence[TimeSeries],
        past_covariates: Optional[Sequence[TimeSeries]],
        future_covariates: Optional[Sequence[TimeSeries]],
        sample_weight: Optional[Sequence[TimeSeries]],
        max_samples_per_ts: Optional[int],
    ) -> DualCovariatesShiftedDataset:
        return DualCovariatesShiftedDataset(
            target_series=target,
            covariates=future_covariates,
            length=self.training_length,
            shift=1,
            max_samples_per_ts=max_samples_per_ts,
            use_static_covariates=self.uses_static_covariates,
            sample_weight=sample_weight,
        )

    def _verify_train_dataset_type(self, train_dataset: TrainingDataset):
        raise_if_not(
            isinstance(train_dataset, DualCovariatesShiftedDataset),
            "RNNModel requires a training dataset of type DualCovariatesShiftedDataset.",
        )
        raise_if_not(
            train_dataset.ds_past.shift == 1,
            "RNNModel requires a shifted training dataset with shift=1.",
        )

    @property
    def supports_multivariate(self) -> bool:
        return True

    @property
    def min_train_series_length(self) -> int:
        return self.training_length + 1

    @property
    def extreme_lags(
        self,
    ) -> tuple[
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        int,
        Optional[int],
    ]:
        return (
            -self.input_chunk_length,
            self.output_chunk_length - 1,
            None,
            None,
            -self.input_chunk_length,
            self.output_chunk_length - 1,
            self.output_chunk_shift,
            self.training_length - self.input_chunk_length,
        )



import inspect
import itertools

import numpy as np
import pandas as pd
import pytest

from darts import TimeSeries
from darts.tests.conftest import TORCH_AVAILABLE
from darts.utils.timeseries_generation import gaussian_timeseries
from darts.utils.utils import freqs

if not TORCH_AVAILABLE:
    pytest.skip(
        f"Torch not available. {__name__} tests will be skipped.",
        allow_module_level=True,
    )

from darts.utils.data import (  # noqa: F401
    DualCovariatesInferenceDataset,
    DualCovariatesSequentialDataset,
    DualCovariatesShiftedDataset,
    FutureCovariatesInferenceDataset,
    FutureCovariatesSequentialDataset,
    FutureCovariatesShiftedDataset,
    HorizonBasedDataset,
    MixedCovariatesInferenceDataset,
    MixedCovariatesSequentialDataset,
    MixedCovariatesShiftedDataset,
    PastCovariatesInferenceDataset,
    PastCovariatesSequentialDataset,
    PastCovariatesShiftedDataset,
    SplitCovariatesInferenceDataset,
    SplitCovariatesSequentialDataset,
    SplitCovariatesShiftedDataset,
)


class TestDataset:
    target1 = gaussian_timeseries(length=100).with_static_covariates(
        pd.Series([0, 1], index=["st1", "st2"])
    )
    target2 = gaussian_timeseries(length=150).with_static_covariates(
        pd.Series([2, 3], index=["st1", "st2"])
    )
    cov_st1 = target1.static_covariates.values
    cov_st2 = target2.static_covariates.values
    cov_st2_df = pd.Series([2, 3], index=["st1", "st2"])
    vals1, vals2 = target1.values(), target2.values()
    cov1, cov2 = gaussian_timeseries(length=100), gaussian_timeseries(length=150)

    def _assert_eq(self, lefts: tuple, rights: tuple):
        for left, right in zip(lefts, rights):
            left = left.values() if isinstance(left, TimeSeries) else left
            right = right.values() if isinstance(right, TimeSeries) else right
            assert type(left) is type(right)
            assert (
                isinstance(
                    left, (TimeSeries, pd.Series, pd.DataFrame, np.ndarray, list)
                )
                or left is None
            )
            if isinstance(left, (pd.Series, pd.DataFrame)):
                assert left.equals(right)
            elif isinstance(left, np.ndarray):
                np.testing.assert_array_equal(left, right)
            elif isinstance(left, (list, TimeSeries)):
                assert left == right
            else:
                assert right is None

    def test_past_covariates_inference_dataset(self):
        # one target series
        ds = PastCovariatesInferenceDataset(
            target_series=self.target1, input_chunk_length=len(self.target1)
        )
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self._assert_eq(ds[0][1:], (None, None, self.cov_st1, self.target1))

        # two target series
        ds = PastCovariatesInferenceDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=max(len(self.target1), len(self.target2)),
        )
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self._assert_eq(ds[1][1:], (None, None, self.cov_st2, self.target2))

        # fail if covariates do not have same size
        with pytest.raises(ValueError):
            ds = PastCovariatesInferenceDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # with covariates
        ds = PastCovariatesInferenceDataset(
            target_series=[self.target1, self.target2],
            covariates=[self.cov1, self.cov2],
            input_chunk_length=max(len(self.target1), len(self.target2)),
        )
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        np.testing.assert_almost_equal(ds[1][1], self.cov2.values())
        self._assert_eq(
            ds[1][2:], (None, self.cov_st2, self.target2)
        )  # no "future past" covariate here

        # more complex case with future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100101", end="20100820", freq="D"
        )  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        short_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = PastCovariatesInferenceDataset(
            target_series=target,
            covariates=short_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # should fail if covariates are too short
        with pytest.raises(ValueError):
            _ = ds[0]

        # Should return correct values when covariates is long enough
        ds = PastCovariatesInferenceDataset(
            target_series=target,
            covariates=long_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], long_cov.values()[-50:-30])
        np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
        assert ds[0][4] == target

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        covariate = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )

        ds = PastCovariatesInferenceDataset(
            target_series=target,
            covariates=covariate,
            input_chunk_length=10,
            output_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], covariate.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], covariate.values()[30:40])
        np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
        assert ds[0][4] == target

    def test_future_covariates_inference_dataset(self):
        # one target series
        ds = FutureCovariatesInferenceDataset(
            target_series=self.target1, input_chunk_length=len(self.target1)
        )
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self._assert_eq(ds[0][1:], (None, self.cov_st1, self.target1))

        # two target series
        ds = FutureCovariatesInferenceDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=max(len(self.target1), len(self.target2)),
        )
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self._assert_eq(ds[1][1:], (None, self.cov_st2, self.target2))

        # fail if covariates do not have same size
        with pytest.raises(ValueError):
            ds = FutureCovariatesInferenceDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # With future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100101", end="20100820", freq="D"
        )  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        short_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = FutureCovariatesInferenceDataset(
            target_series=target, covariates=short_cov, input_chunk_length=10, n=30
        )

        # should fail if covariates are too short
        with pytest.raises(ValueError):
            _ = ds[0]

        # Should return correct values when covariates is long enough
        ds = FutureCovariatesInferenceDataset(
            target_series=target, covariates=long_cov, input_chunk_length=10, n=30
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
        assert ds[0][3] == target

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        covariate = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )

        ds = FutureCovariatesInferenceDataset(
            target_series=target, covariates=covariate, input_chunk_length=10, n=20
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], covariate.values()[30:50])
        np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
        assert ds[0][3] == target

    def test_dual_covariates_inference_dataset(self):
        # one target series
        ds = DualCovariatesInferenceDataset(
            target_series=self.target1, input_chunk_length=len(self.target1)
        )
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self._assert_eq(ds[0][1:], (None, None, self.cov_st1, self.target1))

        # two target series
        ds = DualCovariatesInferenceDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=max(len(self.target1), len(self.target2)),
        )
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self._assert_eq(ds[1][1:], (None, None, self.cov_st2, self.target2))

        # fail if covariates do not have same size
        with pytest.raises(ValueError):
            ds = DualCovariatesInferenceDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # With future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100101", end="20100820", freq="D"
        )  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        short_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = DualCovariatesInferenceDataset(
            target_series=target,
            covariates=short_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # should fail if covariates are too short
        with pytest.raises(ValueError):
            _ = ds[0]

        # Should return correct values when covariates is long enough
        ds = DualCovariatesInferenceDataset(
            target_series=target,
            covariates=long_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], long_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
        assert ds[0][4] == target

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        covariate = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )

        ds = DualCovariatesInferenceDataset(
            target_series=target,
            covariates=covariate,
            input_chunk_length=10,
            output_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], covariate.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], covariate.values()[30:50])
        np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
        assert ds[0][4] == target

    def test_mixed_covariates_inference_dataset(self):
        # With future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100201", end="20100820", freq="D"
        )  # ends 50 days after times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        past_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_past_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )
        future_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = MixedCovariatesInferenceDataset(
            target_series=target,
            past_covariates=past_cov,
            future_covariates=past_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # should fail if future covariates are too short
        with pytest.raises(ValueError):
            _ = ds[0]

        # Should return correct values when covariates is long enough
        ds = MixedCovariatesInferenceDataset(
            target_series=target,
            past_covariates=long_past_cov,
            future_covariates=future_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # It should contain:
        # past_target, past_covariates, historic_future_covariates, future_covariates, future_past_covariates
        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_past_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], future_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][3], future_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][4], long_past_cov.values()[-50:-30])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        assert ds[0][6] == target

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        past_cov = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )
        future_cov = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=30, stop=100, step=1), np.random.randn(70)
        )

        ds = MixedCovariatesInferenceDataset(
            target_series=target,
            past_covariates=past_cov,
            future_covariates=future_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], past_cov.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], future_cov.values()[10:20])
        np.testing.assert_almost_equal(ds[0][3], future_cov.values()[20:40])
        np.testing.assert_almost_equal(ds[0][4], past_cov.values()[30:40])
        np.testing.assert_almost_equal(ds[0][5], self.cov_st2)
        assert ds[0][6] == target

    def test_split_covariates_inference_dataset(self):
        # With future past covariates:
        times1 = pd.date_range(start="20100101", end="20100701", freq="D")
        times2 = pd.date_range(
            start="20100201", end="20100820", freq="D"
        )  # ends 50 days after times1

        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        past_cov = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        )
        long_past_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )
        future_cov = TimeSeries.from_times_and_values(
            times2, np.random.randn(len(times2))
        )

        ds = SplitCovariatesInferenceDataset(
            target_series=target,
            past_covariates=past_cov,
            future_covariates=past_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # should fail if future covariates are too short
        with pytest.raises(ValueError):
            _ = ds[0]

        # Should return correct values when covariates is long enough
        ds = SplitCovariatesInferenceDataset(
            target_series=target,
            past_covariates=long_past_cov,
            future_covariates=future_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=30,
        )

        # It should contain:
        # past_target, past_covariates, future_covariates, future_past_covariates
        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_past_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], future_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][3], long_past_cov.values()[-50:-30])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] == target

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40)
        ).with_static_covariates(self.cov_st2_df)
        past_cov = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60)
        )
        future_cov = TimeSeries.from_times_and_values(
            pd.RangeIndex(start=30, stop=100, step=1), np.random.randn(70)
        )

        ds = SplitCovariatesInferenceDataset(
            target_series=target,
            past_covariates=past_cov,
            future_covariates=future_cov,
            input_chunk_length=10,
            output_chunk_length=10,
            n=20,
        )

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], past_cov.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], future_cov.values()[20:40])
        np.testing.assert_almost_equal(ds[0][3], past_cov.values()[30:40])
        np.testing.assert_almost_equal(ds[0][4], self.cov_st2)
        assert ds[0][5] == target

    @pytest.mark.parametrize(
        "config",
        [
            # (dataset class, whether contains future, future batch index)
            (PastCovariatesInferenceDataset, None),
            (FutureCovariatesInferenceDataset, 1),
            (DualCovariatesInferenceDataset, 2),
            (MixedCovariatesInferenceDataset, 3),
            (SplitCovariatesInferenceDataset, 2),
        ],
    )
    def test_inference_dataset_output_chunk_shift(self, config):
        ds_cls, future_idx = config
        ocl = 1
        ocs = 2
        target = self.target1[: -(ocl + ocs)]

        ds_covs = {}
        ds_init_params = set(inspect.signature(ds_cls.__init__).parameters)
        for cov_type in ["covariates", "past_covariates", "future_covariates"]:
            if cov_type in ds_init_params:
                ds_covs[cov_type] = self.cov1

        with pytest.raises(ValueError) as err:
            _ = ds_cls(
                target_series=target,
                input_chunk_length=1,
                output_chunk_length=1,
                output_chunk_shift=1,
                n=2,
                **ds_covs,
            )
        assert str(err.value).startswith("Cannot perform auto-regression")

        # regular dataset with output shift=0 and ocl=3: the 3rd future values should be identical to the 1st future
        # values of a dataset with output shift=2 and ocl=1
        ds_reg = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            output_chunk_shift=0,
            n=1,
            **ds_covs,
        )

        ds_shift = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=1,
            output_chunk_shift=ocs,
            n=1,
            **ds_covs,
        )

        batch_reg, batch_shift = ds_reg[0], ds_shift[0]

        # shifted prediction starts 2 steps after regular prediction
        assert batch_reg[-1] == batch_shift[-1] - ocs * target.freq

        if future_idx is not None:
            # 3rd future values of regular ds must be identical to the 1st future values of shifted dataset
            np.testing.assert_array_equal(
                batch_reg[future_idx][ocs:], batch_shift[future_idx]
            )
            batch_reg = batch_reg[:future_idx] + batch_reg[future_idx + 1 :]
            batch_shift = batch_shift[:future_idx] + batch_shift[future_idx + 1 :]

        # without future part, the input will be identical between regular, and shifted dataset
        assert all([
            np.all(el_reg == el_shift)
            for el_reg, el_shift in zip(batch_reg[:-1], batch_shift[:-1])
        ])

    def test_past_covariates_sequential_dataset(self):
        # one target series
        ds = PastCovariatesSequentialDataset(
            target_series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 81
        self._assert_eq(
            ds[5], (self.target1[75:85], None, self.cov_st1, None, self.target1[85:95])
        )

        # two target series
        ds = PastCovariatesSequentialDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 262
        self._assert_eq(
            ds[5], (self.target1[75:85], None, self.cov_st1, None, self.target1[85:95])
        )
        self._assert_eq(
            ds[136],
            (self.target2[125:135], None, self.cov_st2, None, self.target2[135:145]),
        )

        # two target series with custom max_nr_samples
        ds = PastCovariatesSequentialDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5], (self.target1[75:85], None, self.cov_st1, None, self.target1[85:95])
        )
        self._assert_eq(
            ds[55],
            (self.target2[125:135], None, self.cov_st2, None, self.target2[135:145]),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = PastCovariatesSequentialDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # two targets and two covariates
        ds = PastCovariatesSequentialDataset(
            target_series=[self.target1, self.target2],
            covariates=[self.cov1, self.cov2],
            input_chunk_length=10,
            output_chunk_length=10,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[75:85],
                self.cov1[75:85],
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[136],
            (
                self.target2[125:135],
                self.cov2[125:135],
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # should fail if covariates do not have the required time span, even though covariates are longer
        times1 = pd.date_range(start="20100101", end="20110101", freq="D")
        times2 = pd.date_range(start="20120101", end="20150101", freq="D")
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesSequentialDataset(
            target_series=target,
            covariates=cov,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        with pytest.raises(ValueError):
            _ = ds[5]

        # the same should fail when series are integer-indexed
        times1 = pd.RangeIndex(start=0, stop=100, step=1)
        times2 = pd.RangeIndex(start=200, stop=400, step=1)
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesSequentialDataset(
            target_series=target,
            covariates=cov,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        with pytest.raises(ValueError):
            _ = ds[5]

        # we should get the correct covariate slice even when target and covariates are not aligned
        times1 = pd.date_range(start="20100101", end="20110101", freq="D")
        times2 = pd.date_range(start="20090101", end="20110106", freq="D")
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesSequentialDataset(
            target_series=target,
            covariates=cov,
            input_chunk_length=10,
            output_chunk_length=10,
        )

        np.testing.assert_almost_equal(ds[0][1], cov.values()[-25:-15])
        np.testing.assert_almost_equal(ds[5][1], cov.values()[-30:-20])

        # This should also be the case when series are integer indexed
        times1 = pd.RangeIndex(start=100, stop=200, step=1)
        times2 = pd.RangeIndex(start=50, stop=250, step=1)
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesSequentialDataset(
            target_series=target,
            covariates=cov,
            input_chunk_length=10,
            output_chunk_length=10,
        )

        np.testing.assert_almost_equal(ds[0][1], cov.values()[-70:-60])
        np.testing.assert_almost_equal(ds[5][1], cov.values()[-75:-65])

    def test_future_covariates_sequential_dataset(self):
        # one target series
        ds = FutureCovariatesSequentialDataset(
            target_series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 81
        self._assert_eq(
            ds[5], (self.target1[75:85], None, self.cov_st1, None, self.target1[85:95])
        )

        # two target series
        ds = FutureCovariatesSequentialDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 262
        self._assert_eq(
            ds[5], (self.target1[75:85], None, self.cov_st1, None, self.target1[85:95])
        )
        self._assert_eq(
            ds[136],
            (self.target2[125:135], None, self.cov_st2, None, self.target2[135:145]),
        )

        # two target series with custom max_nr_samples
        ds = FutureCovariatesSequentialDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5], (self.target1[75:85], None, self.cov_st1, None, self.target1[85:95])
        )
        self._assert_eq(
            ds[55],
            (self.target2[125:135], None, self.cov_st2, None, self.target2[135:145]),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = FutureCovariatesSequentialDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # two targets and two covariates; covariates not aligned, must contain correct values
        target1 = TimeSeries.from_values(np.random.randn(100)).with_static_covariates(
            self.cov_st2_df
        )
        target2 = TimeSeries.from_values(np.random.randn(50)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(120))
        cov2 = TimeSeries.from_values(np.random.randn(80))

        ds = FutureCovariatesSequentialDataset(
            target_series=[target1, target2],
            covariates=[cov1, cov2],
            input_chunk_length=10,
            output_chunk_length=10,
        )

        np.testing.assert_almost_equal(ds[0][0], target1.values()[-20:-10])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-30:-20])
        np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[0][4], target1.values()[-10:])

        np.testing.assert_almost_equal(ds[101][0], target2.values()[-40:-30])
        np.testing.assert_almost_equal(ds[101][1], cov2.values()[-60:-50])
        np.testing.assert_almost_equal(ds[101][2], self.cov_st2)
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[101][4], target2.values()[-30:-20])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = FutureCovariatesSequentialDataset(
            target_series=[target1],
            covariates=[cov1],
            input_chunk_length=2,
            output_chunk_length=2,
        )

        np.testing.assert_almost_equal(ds[0][0], target1.values()[-4:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-4:-2])
        np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[0][4], target1.values()[-2:])

        # Should fail if covariates are not long enough
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(7))

        ds = FutureCovariatesSequentialDataset(
            target_series=[target1],
            covariates=[cov1],
            input_chunk_length=2,
            output_chunk_length=2,
        )

        with pytest.raises(ValueError):
            _ = ds[0]

    def test_dual_covariates_sequential_dataset(self):
        # Must contain (past_target, historic_future_covariates, future_covariates, static covariates,
        # sample weight, future_target)

        # one target series
        ds = DualCovariatesSequentialDataset(
            target_series=self.target1,
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 81
        self._assert_eq(
            ds[5],
            (self.target1[75:85], None, None, self.cov_st1, None, self.target1[85:95]),
        )

        # two target series
        ds = DualCovariatesSequentialDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
        )
        assert len(ds) == 262
        self._assert_eq(
            ds[5],
            (self.target1[75:85], None, None, self.cov_st1, None, self.target1[85:95]),
        )
        self._assert_eq(
            ds[136],
            (
                self.target2[125:135],
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two target series with custom max_nr_samples
        ds = DualCovariatesSequentialDataset(
            target_series=[self.target1, self.target2],
            input_chunk_length=10,
            output_chunk_length=10,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5],
            (self.target1[75:85], None, None, self.cov_st1, None, self.target1[85:95]),
        )
        self._assert_eq(
            ds[55],
            (
                self.target2[125:135],
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = DualCovariatesSequentialDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # two targets and two covariates; covariates not aligned, must contain correct values
        target1 = TimeSeries.from_values(np.random.randn(100)).with_static_covariates(
            self.cov_st2_df
        )
        target2 = TimeSeries.from_values(np.random.randn(50)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(120))
        cov2 = TimeSeries.from_values(np.random.randn(80))

        ds = DualCovariatesSequentialDataset(
            target_series=[target1, target2],
            covariates=[cov1, cov2],
            input_chunk_length=10,
            output_chunk_length=10,
        )

        np.testing.assert_almost_equal(ds[0][0], target1.values()[-20:-10])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-40:-30])
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-30:-20])
        np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
        assert ds[0][4] is None
        np.testing.assert_almost_equal(ds[0][5], target1.values()[-10:])

        np.testing.assert_almost_equal(ds[101][0], target2.values()[-40:-30])
        np.testing.assert_almost_equal(ds[101][1], cov2.values()[-70:-60])
        np.testing.assert_almost_equal(ds[101][2], cov2.values()[-60:-50])
        np.testing.assert_almost_equal(ds[101][3], self.cov_st2)
        assert ds[101][4] is None
        np.testing.assert_almost_equal(ds[101][5], target2.values()[-30:-20])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = DualCovariatesSequentialDataset(
            target_series=[target1],
            covariates=[cov1],
            input_chunk_length=2,
            output_chunk_length=2,
        )

        np.testing.assert_almost_equal(ds[0][0], target1.values()[-4:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-6:-4])
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-4:-2])
        np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
        assert ds[0][4] is None
        np.testing.assert_almost_equal(ds[0][5], target1.values()[-2:])

        # Should fail if covariates are not long enough
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(7))

        ds = DualCovariatesSequentialDataset(
            target_series=[target1],
            covariates=[cov1],
            input_chunk_length=2,
            output_chunk_length=2,
        )

        with pytest.raises(ValueError):
            _ = ds[0]

    def test_past_covariates_shifted_dataset(self):
        # one target series
        ds = PastCovariatesShiftedDataset(
            target_series=self.target1, length=10, shift=5
        )
        assert len(ds) == 86
        self._assert_eq(
            ds[5], (self.target1[80:90], None, self.cov_st1, None, self.target1[85:95])
        )

        # two target series
        ds = PastCovariatesShiftedDataset(
            target_series=[self.target1, self.target2], length=10, shift=5
        )
        assert len(ds) == 272
        self._assert_eq(
            ds[5], (self.target1[80:90], None, self.cov_st1, None, self.target1[85:95])
        )
        self._assert_eq(
            ds[141],
            (self.target2[130:140], None, self.cov_st2, None, self.target2[135:145]),
        )

        # two target series with custom max_nr_samples
        ds = PastCovariatesShiftedDataset(
            target_series=[self.target1, self.target2],
            length=10,
            shift=5,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5], (self.target1[80:90], None, self.cov_st1, None, self.target1[85:95])
        )
        self._assert_eq(
            ds[55],
            (self.target2[130:140], None, self.cov_st2, None, self.target2[135:145]),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = PastCovariatesShiftedDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # two targets and two covariates
        ds = PastCovariatesShiftedDataset(
            target_series=[self.target1, self.target2],
            covariates=[self.cov1, self.cov2],
            length=10,
            shift=5,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                self.cov1[80:90],
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                self.cov2[130:140],
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = PastCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
        np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[0][4], target1.values()[-3:])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
        np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[0][4], target1.values()[-3:])

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(5))
        ds = PastCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        with pytest.raises(ValueError):
            _ = ds[0]

    def test_future_covariates_shifted_dataset(self):
        # one target series
        ds = FutureCovariatesShiftedDataset(
            target_series=self.target1, length=10, shift=5
        )
        assert len(ds) == 86
        self._assert_eq(
            ds[5], (self.target1[80:90], None, self.cov_st1, None, self.target1[85:95])
        )

        # two target series
        ds = FutureCovariatesShiftedDataset(
            target_series=[self.target1, self.target2], length=10, shift=5
        )
        assert len(ds) == 272
        self._assert_eq(
            ds[5], (self.target1[80:90], None, self.cov_st1, None, self.target1[85:95])
        )
        self._assert_eq(
            ds[141],
            (self.target2[130:140], None, self.cov_st2, None, self.target2[135:145]),
        )

        # two target series with custom max_nr_samples
        ds = FutureCovariatesShiftedDataset(
            target_series=[self.target1, self.target2],
            length=10,
            shift=5,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5], (self.target1[80:90], None, self.cov_st1, None, self.target1[85:95])
        )
        self._assert_eq(
            ds[55],
            (self.target2[130:140], None, self.cov_st2, None, self.target2[135:145]),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = FutureCovariatesShiftedDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # two targets and two covariates
        ds = FutureCovariatesShiftedDataset(
            target_series=[self.target1, self.target2],
            covariates=[self.cov1, self.cov2],
            length=10,
            shift=5,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                self.cov1[85:95],
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                self.cov2[135:145],
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = FutureCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[0][4], target1.values()[-3:])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = FutureCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][2], self.cov_st2)
        assert ds[0][3] is None
        np.testing.assert_almost_equal(ds[0][4], target1.values()[-3:])

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(7))
        ds = FutureCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        with pytest.raises(ValueError):
            _ = ds[0]

    def test_dual_covariates_shifted_dataset(self):
        # one target series
        ds = DualCovariatesShiftedDataset(
            target_series=self.target1, length=10, shift=5
        )
        assert len(ds) == 86
        self._assert_eq(
            ds[5],
            (self.target1[80:90], None, None, self.cov_st1, None, self.target1[85:95]),
        )

        # two target series
        ds = DualCovariatesShiftedDataset(
            target_series=[self.target1, self.target2], length=10, shift=5
        )
        assert len(ds) == 272
        self._assert_eq(
            ds[5],
            (self.target1[80:90], None, None, self.cov_st1, None, self.target1[85:95]),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two target series with custom max_nr_samples
        ds = DualCovariatesShiftedDataset(
            target_series=[self.target1, self.target2],
            length=10,
            shift=5,
            max_samples_per_ts=50,
        )
        assert len(ds) == 100
        self._assert_eq(
            ds[5],
            (self.target1[80:90], None, None, self.cov_st1, None, self.target1[85:95]),
        )
        self._assert_eq(
            ds[55],
            (
                self.target2[130:140],
                None,
                None,
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = DualCovariatesShiftedDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # two targets and two covariates
        ds = DualCovariatesShiftedDataset(
            target_series=[self.target1, self.target2],
            covariates=[self.cov1, self.cov2],
            length=10,
            shift=5,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[80:90],
                self.cov1[80:90],
                self.cov1[85:95],
                self.cov_st1,
                None,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[141],
            (
                self.target2[130:140],
                self.cov2[130:140],
                self.cov2[135:145],
                self.cov_st2,
                None,
                self.target2[135:145],
            ),
        )

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = DualCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
        assert ds[0][4] is None
        np.testing.assert_almost_equal(ds[0][5], target1.values()[-3:])

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start="20090201", end="20090220", freq="D")
        times2 = pd.date_range(start="20090201", end="20090222", freq="D")
        target1 = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = DualCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        np.testing.assert_almost_equal(ds[0][0], target1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][1], cov1.values()[-7:-4])
        np.testing.assert_almost_equal(ds[0][2], cov1.values()[-5:-2])
        np.testing.assert_almost_equal(ds[0][3], self.cov_st2)
        assert ds[0][4] is None
        np.testing.assert_almost_equal(ds[0][5], target1.values()[-3:])

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8)).with_static_covariates(
            self.cov_st2_df
        )
        cov1 = TimeSeries.from_values(np.random.randn(7))
        ds = DualCovariatesShiftedDataset(
            target_series=[target1], covariates=[cov1], length=3, shift=2
        )
        with pytest.raises(ValueError):
            _ = ds[0]

    @pytest.mark.parametrize("use_weight", [False, True])
    def test_horizon_based_dataset(self, use_weight):
        weight1 = self.target1 + 1
        weight2 = self.target2 + 1

        weight = weight1 if use_weight else None
        weight_exp = weight1[85:95] if use_weight else None
        # one target series
        ds = HorizonBasedDataset(
            target_series=self.target1,
            output_chunk_length=10,
            lh=(1, 3),
            lookback=2,
            sample_weight=weight,
        )
        assert len(ds) == 20
        self._assert_eq(
            ds[5],
            (self.target1[65:85], None, self.cov_st1, weight_exp, self.target1[85:95]),
        )

        # two target series
        weight = [weight1, weight2] if use_weight else None
        weight_exp1 = weight1[85:95] if use_weight else None
        weight_exp2 = weight2[135:145] if use_weight else None
        ds = HorizonBasedDataset(
            target_series=[self.target1, self.target2],
            output_chunk_length=10,
            lh=(1, 3),
            lookback=2,
            sample_weight=weight,
        )
        assert len(ds) == 40
        self._assert_eq(
            ds[5],
            (self.target1[65:85], None, self.cov_st1, weight_exp1, self.target1[85:95]),
        )
        self._assert_eq(
            ds[25],
            (
                self.target2[115:135],
                None,
                self.cov_st2,
                weight_exp2,
                self.target2[135:145],
            ),
        )

        # two targets and one covariate
        with pytest.raises(ValueError):
            ds = HorizonBasedDataset(
                target_series=[self.target1, self.target2], covariates=[self.cov1]
            )

        # two targets and two covariates
        weight = [weight1, weight2] if use_weight else None
        weight_exp1 = weight1[85:95] if use_weight else None
        weight_exp2 = weight2[135:145] if use_weight else None
        ds = HorizonBasedDataset(
            target_series=[self.target1, self.target2],
            covariates=[self.cov1, self.cov2],
            output_chunk_length=10,
            lh=(1, 3),
            lookback=2,
            sample_weight=weight,
        )
        self._assert_eq(
            ds[5],
            (
                self.target1[65:85],
                self.cov1[65:85],
                self.cov_st1,
                weight_exp1,
                self.target1[85:95],
            ),
        )
        self._assert_eq(
            ds[25],
            (
                self.target2[115:135],
                self.cov2[115:135],
                self.cov_st2,
                weight_exp2,
                self.target2[135:145],
            ),
        )

    @pytest.mark.parametrize(
        "config",
        [
            # (dataset class, whether contains future, future batch index)
            (PastCovariatesSequentialDataset, None),
            (FutureCovariatesSequentialDataset, 1),
            (DualCovariatesSequentialDataset, 2),
            (MixedCovariatesSequentialDataset, 3),
            (SplitCovariatesSequentialDataset, 2),
        ],
    )
    def test_sequential_training_dataset_output_chunk_shift(self, config):
        ds_cls, future_idx = config
        ocl = 1
        ocs = 2
        target = self.target1[: -(ocl + ocs)]
        sample_weight = target + 1

        ds_covs = {}
        ds_init_params = set(inspect.signature(ds_cls.__init__).parameters)
        for cov_type in ["covariates", "past_covariates", "future_covariates"]:
            if cov_type in ds_init_params:
                ds_covs[cov_type] = self.cov1

        # regular dataset with output shift=0 and ocl=3: the 3rd future values should be identical to the 1st future
        # values of a dataset with output shift=2 and ocl=1
        ds_reg = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            output_chunk_shift=0,
            sample_weight=sample_weight,
            **ds_covs,
        )

        ds_shift = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=1,
            output_chunk_shift=ocs,
            sample_weight=sample_weight,
            **ds_covs,
        )

        batch_reg, batch_shift = ds_reg[0], ds_shift[0]

        if future_idx is not None:
            # 3rd future values of regular ds must be identical to the 1st future values of shifted dataset
            np.testing.assert_array_equal(
                batch_reg[future_idx][-1:], batch_shift[future_idx]
            )
            batch_reg = batch_reg[:future_idx] + batch_reg[future_idx + 1 :]
            batch_shift = batch_shift[:future_idx] + batch_shift[future_idx + 1 :]

        # last two elements are (sample weight, output chunk of the target series).
        # 3rd future values of regular ds must be identical to the 1st future values of shifted dataset
        batch_reg = batch_reg[:-2] + (batch_reg[-2][ocs:], batch_reg[-1][ocs:])

        # without future part, the input will be identical between regular, and shifted dataset
        assert all([
            np.all(el_reg == el_shift)
            for el_reg, el_shift in zip(batch_reg[:-1], batch_shift[:-1])
        ])

    @pytest.mark.parametrize(
        "config",
        itertools.product(
            [
                PastCovariatesSequentialDataset,
                FutureCovariatesSequentialDataset,
                DualCovariatesSequentialDataset,
                MixedCovariatesSequentialDataset,
                SplitCovariatesSequentialDataset,
            ],
            [True, False],
        ),
    )
    def test_sequential_training_dataset_weight(self, config):
        ds_cls, manual_weight = config

        def get_built_in_weigths(targets):
            if isinstance(targets, list):
                max_steps = max([len(ts) for ts in targets])
            else:
                max_steps = len(targets)
            weight_expected = np.linspace(0, 1, max_steps)[-3:]
            return np.expand_dims(weight_expected, -1)

        target1 = self.target1
        target2 = self.target2
        weight1 = target1 + 1
        weight2 = target2 + 1
        built_in_weight = "linear"

        ds_covs = {}
        ds_init_params = set(inspect.signature(ds_cls.__init__).parameters)
        for cov_type in ["covariates", "past_covariates", "future_covariates"]:
            if cov_type in ds_init_params:
                ds_covs[cov_type] = self.cov1

        # no sample weight
        ds = ds_cls(
            target_series=target1,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=None,
            **ds_covs,
        )
        assert ds[0][-2] is None

        # whenever we use sample weight, the weight are extracted from the same time frame as the target labels
        # since we set the weight to be `target + 1`, the returned batch weight must also be `batch_target_label + 1`

        # single univariate
        target = target1
        weight = weight1 if manual_weight else built_in_weight
        ds = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # single univariate with longer weight
        target = target1
        weight = (
            weight1.prepend_values([0.0]).append_values([0.0])
            if manual_weight
            else built_in_weight
        )
        ds = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # single multivariate with multivariate weight
        target = target1.stack(target1 + 1)
        weight = weight1.stack(weight1 + 1) if manual_weight else built_in_weight
        ds = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # single multivariate with univariate (global) weight
        target = target1.stack(target1 + 1)
        weight = weight1 if manual_weight else built_in_weight
        ds = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        # output weight corresponds to first target component + 1 (e.g. weight1)
        weight_exp = (
            ds[0][-1][:, 0:1] + 1 if manual_weight else get_built_in_weigths(target)
        )
        assert np.all(ds[0][-2] == weight_exp)

        # single univariate and list of single weight
        target = target1
        weight = [weight1] if manual_weight else built_in_weight
        ds = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **ds_covs,
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # multiple univariate
        target = [target1, target2]
        weight = [weight1, weight2] if manual_weight else built_in_weight
        ds = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **{k: [v] * 2 for k, v in ds_covs.items()},
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

        # multiple multivariate
        target = [target1.stack(target1 + 1), target2.stack(target2 + 1)]
        weight = (
            [weight1.stack(weight1 + 1), weight2.stack(weight2 + 1)]
            if manual_weight
            else built_in_weight
        )
        ds = ds_cls(
            target_series=target,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=weight,
            **{k: [v] * 2 for k, v in ds_covs.items()},
        )
        weight_exp = ds[0][-1] + 1 if manual_weight else get_built_in_weigths(target)
        assert np.all(ds[0][-2] == weight_exp)

    @pytest.mark.parametrize(
        "ds_cls",
        [
            PastCovariatesSequentialDataset,
            FutureCovariatesSequentialDataset,
            DualCovariatesSequentialDataset,
            MixedCovariatesSequentialDataset,
            SplitCovariatesSequentialDataset,
        ],
    )
    def test_sequential_training_dataset_invalid_weight(self, ds_cls):
        ts = self.target1

        # invalid built-in weight
        with pytest.raises(ValueError) as err:
            _ = ds_cls(
                target_series=[ts, ts],
                input_chunk_length=1,
                output_chunk_length=3,
                sample_weight="invalid",
            )
        assert str(err.value).startswith(
            "Invalid `sample_weight` value: `'invalid'`. If a string, must be one of: "
        )

        # mismatch number of target and weight series
        with pytest.raises(ValueError) as err:
            _ = ds_cls(
                target_series=[ts, ts],
                input_chunk_length=1,
                output_chunk_length=3,
                sample_weight=[ts],
            )
        assert (
            str(err.value)
            == "The provided sequence of target `series` must have the same "
            "length as the provided sequence of `sample_weight`."
        )

        # too many weight components
        ds = ds_cls(
            target_series=ts,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=ts.stack(ts + 1),
        )
        with pytest.raises(ValueError) as err:
            _ = ds[0]
        assert (
            str(err.value)
            == "The number of components in `sample_weight` must either be `1` or match "
            "the number of target series components `1`. (0-th series)"
        )

        # weight too short end
        ds = ds_cls(
            target_series=ts,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=ts[:-1],
        )
        with pytest.raises(ValueError) as err:
            _ = ds[0]
        assert (
            str(err.value)
            == "Missing sample weights; could not find sample weights in index value range: "
            "2000-04-07 00:00:00 - 2000-04-09 00:00:00."
        )

        # weight too short start
        ds = ds_cls(
            target_series=ts,
            input_chunk_length=1,
            output_chunk_length=3,
            sample_weight=ts[2:],
        )
        with pytest.raises(ValueError) as err:
            _ = ds[len(ds) - 1]
        assert (
            str(err.value)
            == "Missing sample weights; could not find sample weights in index value range: "
            "2000-01-02 00:00:00 - 2000-01-04 00:00:00."
        )

    def test_get_matching_index(self):
        from darts.utils.data.utils import _get_matching_index

        # Check dividable freq
        times1 = pd.date_range(start="20100101", end="20100330", freq="D")
        times2 = pd.date_range(start="20100101", end="20100320", freq="D")
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        assert _get_matching_index(target, cov, idx=15) == 5

        # check non-dividable freq
        times1 = pd.date_range(start="20100101", end="20120101", freq=freqs["ME"])
        times2 = pd.date_range(start="20090101", end="20110601", freq=freqs["ME"])
        target = TimeSeries.from_times_and_values(
            times1, np.random.randn(len(times1))
        ).with_static_covariates(self.cov_st2_df)
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        assert _get_matching_index(target, cov, idx=15) == 15 - 7

        # check integer-indexed series
        times2 = pd.RangeIndex(start=10, stop=90)
        target = TimeSeries.from_values(np.random.randn(100)).with_static_covariates(
            self.cov_st2_df
        )
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        assert _get_matching_index(target, cov, idx=15) == 5



"""
NLL Exponential Scorer
----------------------

Exponential distribution negative log-likelihood Scorer.

The anomaly score is the negative log likelihood of the actual series values
under an Exponential distribution estimated from the stochastic prediction.
"""

import numpy as np
from scipy.stats import expon

from darts.ad.scorers.scorers import NLLScorer


class ExponentialNLLScorer(NLLScorer):
    def __init__(self, window: int = 1) -> None:
        """NLL Exponential Scorer

        Parameters
        ----------
        window
            Integer value indicating the size of the window W used by the scorer to transform the series into an
            anomaly score. A scorer will slice the given series into subsequences of size W and returns a value
            indicating how anomalous these subset of W values are. A post-processing step will convert this anomaly
            score into a point-wise anomaly score (see definition of `window_transform`). The window size should be
            commensurate to the expected durations of the anomalies one is looking for.
        """
        super().__init__(window=window)

    def __str__(self):
        return "ExponentialNLLScorer"

    def _score_core_nllikelihood(
        self, vals: np.ndarray, pred_vals: np.ndarray
    ) -> np.ndarray:
        # This is the ML estimate for 1/lambda, which is what scipy expects as scale.
        mu = np.mean(pred_vals, axis=1)
        # This is ML estimate for the loc - see:
        # https://github.com/scipy/scipy/blob/de80faf9d3480b9dbb9b888568b64499e0e70c19/scipy/stats/_continuous_distns.py#L1705
        loc = np.min(pred_vals, axis=1)
        return -expon.logpdf(vals, scale=mu, loc=loc)



"""
Anomaly models base classes
"""

import sys
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Literal, Optional, Union

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from darts.ad.scorers.scorers import AnomalyScorer
from darts.ad.utils import (
    _assert_same_length,
    _check_input,
    eval_metric_from_scores,
    show_anomalies_from_scores,
)
from darts.logging import get_logger, raise_log
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class AnomalyModel(ABC):
    """Base class for all anomaly models."""

    def __init__(self, model, scorer):
        self.scorers = [scorer] if not isinstance(scorer, Sequence) else scorer
        if not all([isinstance(s, AnomalyScorer) for s in self.scorers]):
            raise_log(
                ValueError(
                    "all scorers must be of instance `darts.ad.scorers.AnomalyScorer`."
                ),
                logger=logger,
            )
        self.model = model

    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        allow_model_training: bool,
        **kwargs,
    ) -> Self:
        """Fit the underlying forecasting/filtering model (if applicable) and the fittable scorers."""
        # interrupt training if nothing to fit
        if not allow_model_training and not self.scorers_are_trainable:
            return self

        # check input series and covert to sequences
        series, kwargs = self._process_input_series(series, **kwargs)
        self._fit_core(
            series=series, allow_model_training=allow_model_training, **kwargs
        )
        return self

    @abstractmethod
    def score(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        return_model_prediction: bool = False,
        **kwargs,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Compute anomaly score(s) for the given series.

        Predicts the given target time series with the forecasting model, and applies the scorer(s)
        on the prediction and the target input time series.

        Parameters
        ----------
        series
            The (sequence of) series to score on.
        return_model_prediction
            Whether to return the forecasting/filtering model prediction along with the anomaly scores.
        **kwargs
            Additional parameters passed to `AnomalyModel.predict_series()`

        Returns
        -------
        TimeSeries
            A single `TimeSeries` for a single `series` with a single anomaly scorers.
        Sequence[TimeSeries]
            A sequence of `TimeSeries` for:

            - a single `series` with multiple anomaly scorers.
            - a sequence of `series` with a single anomaly scorer.
        Sequence[Sequence[TimeSeries]]
            A sequence of sequences of `TimeSeries` for a sequence of `series` and multiple anomaly scorers.
            The outer sequence is over the series, and inner sequence is over the scorers.
        """
        called_with_single_series = isinstance(series, TimeSeries)
        # check input series and covert to sequences
        series, kwargs = self._process_input_series(series, **kwargs)
        # predict / filter `series`
        pred = self.predict_series(series=series, **kwargs)

        scores = list(
            zip(*[sc.score_from_prediction(series, pred) for sc in self.scorers])
        )

        if called_with_single_series:
            scores = scores[0]
            if len(scores) == 1:
                # there's only one scorer
                scores = scores[0]
            pred = pred[0]

        if return_model_prediction:
            return scores, pred

        return scores

    @abstractmethod
    def predict_series(
        self, series: Sequence[TimeSeries], **kwargs
    ) -> Sequence[TimeSeries]:
        """Abstract method to implement the generation of predictions for the input `series`."""
        pass

    def eval_metric(
        self,
        anomalies: Union[TimeSeries, Sequence[TimeSeries]],
        series: Union[TimeSeries, Sequence[TimeSeries]],
        metric: Literal["AUC_ROC", "AUC_PR"] = "AUC_ROC",
        **kwargs,
    ) -> Union[
        dict[str, float],
        dict[str, Sequence[float]],
        Sequence[dict[str, float]],
        Sequence[dict[str, Sequence[float]]],
    ]:
        """Compute the accuracy of the anomaly scores computed by the model.

        Predicts the `series` with the underlying forecasting/filtering model, and applies the scorer(s) on the
        predicted time series and the given target time series. Returns the score(s) of an agnostic threshold metric,
        based on the anomaly score given by the scorer(s).

        Parameters
        ----------
        anomalies
            The (sequence of) ground truth binary anomaly series (`1` if it is an anomaly and `0` if not).
        series
            The (sequence of) series to predict anomalies on.
        metric
            The name of the metric function to use. Must be one of "AUC_ROC" (Area Under the
            Receiver Operating Characteristic Curve) and "AUC_PR" (Average Precision from scores).
            Default: "AUC_ROC".
        **kwargs
            Additional parameters passed to the `score()` method.

        Returns
        -------
        Dict[str, float]
            A dictionary with the resulting metrics for single univariate `series`, with keys representing the
            anomaly scorer(s), and values representing the metric values.
        Dict[str, Sequence[float]]
            Same as for `Dict[str, float]` but for multivariate `series`, and anomaly scorers that treat series
            components/columns independently (by nature of the scorer or if `component_wise=True`).
        Sequence[Dict[str, float]]
            Same as for `Dict[str, float]` but for a sequence of univariate series.
        Sequence[Dict[str, Sequence[float]]]
            Same as for `Dict[str, float]` but for a sequence of multivariate series.
        """

        def _check_univariate(s: TimeSeries):
            """Checks if `anomalies` contains only univariate series, which
            is required if any of the scorers returns a univariate score.
            """
            if self.scorers_are_univariate and not s.width == 1:
                raise_log(
                    ValueError(
                        f"Anomaly model contains scorer {[s.__str__() for s in self.scorers if s.is_univariate]} "
                        f"that will return a univariate anomaly score series (width=1). Found a multivariate "
                        f"`anomalies`. The evaluation of the accuracy cannot be computed. If applicable, "
                        f"think about setting the scorer parameter `componenet_wise` to True."
                    ),
                    logger=logger,
                )

        called_with_single_series = isinstance(series, TimeSeries)
        # deterministic `series`
        series = _check_input(
            series,
            name="series",
            check_deterministic=True,
        )
        # deterministic, binary anomalies, (possibly univariate)
        anomalies = _check_input(
            anomalies,
            name="anomalies",
            check_deterministic=True,
            check_binary=True,
            extra_checks=_check_univariate,
        )
        _assert_same_length(series, anomalies, "series", "anomalies")

        pred_scores = self.score(series=series, **kwargs)

        # compute metric for anomaly scores
        windows = [s.window for s in self.scorers]

        # create a list of unique names for each scorer that
        # will be used as keys for the dictionary containing
        # the accuracy of each scorer.
        name_scorers = []
        for scorer in self.scorers:
            name = scorer.__str__() + "_w=" + str(scorer.window)

            if name in name_scorers:
                i = 1
                new_name = name + "_" + str(i)
                while new_name in name_scorers:
                    i = i + 1
                    new_name = name + "_" + str(i)
                name = new_name

            name_scorers.append(name)

        metric_vals = []
        for anomalies, scores in zip(anomalies, pred_scores):
            metric_vals.append(
                eval_metric_from_scores(
                    anomalies=anomalies,
                    pred_scores=scores,
                    window=windows,
                    metric=metric,
                )
            )
        metric_vals_pred_scores = [
            dict(zip(name_scorers, scorer_values)) for scorer_values in metric_vals
        ]

        return (
            metric_vals_pred_scores[0]
            if called_with_single_series
            else metric_vals_pred_scores
        )

    def show_anomalies(
        self,
        series: TimeSeries,
        anomalies: TimeSeries = None,
        predict_kwargs: Optional[dict] = None,
        names_of_scorers: Union[str, Sequence[str]] = None,
        title: str = None,
        metric: Optional[Literal["AUC_ROC", "AUC_PR"]] = None,
        component_wise: bool = False,
        **score_kwargs,
    ):
        """Plot the results of the anomaly model.

        Computes the score on the given series input and shows the different anomaly scores with respect to time.

        The plot will be composed of the following:

        - the series itself with the output of the forecasting model.
        - the anomaly score for each scorer. The scorers with different windows will be separated.
        - the actual anomalies, if given.

        It is possible to:

        - add a title to the figure with the parameter `title`
        - give personalized names for the scorers with `names_of_scorers`
        - show the results of a metric for each anomaly score (AUC_ROC or AUC_PR), if the actual anomalies are provided.

        Parameters
        ----------
        series
            The series to visualize anomalies from.
        anomalies
            The ground truth of the anomalies (1 if it is an anomaly and 0 if not).
        predict_kwargs
            Optionally, some additional parameters passed to `AnomalyModel.predict_series()`.
        names_of_scorers
            Name of the scores. Must be a list of length equal to the number of scorers in the anomaly_model.
        title
            Title of the figure.
        metric
            Optionally, the name of the metric function to use. Must be one of "AUC_ROC" (Area Under the
            Receiver Operating Characteristic Curve) and "AUC_PR" (Average Precision from scores).
            Default: "AUC_ROC".
        score_kwargs
            parameters for the `score()` method.
        component_wise
            If True, will separately plot each component in case of multivariate anomaly detection.
        """
        series = _check_input(series, name="series", num_series_expected=1)[0]
        predict_kwargs = predict_kwargs if predict_kwargs is not None else {}
        pred_scores, pred_series = self.score(
            series,
            return_model_prediction=True,
            **predict_kwargs,
            **score_kwargs,
        )

        if title is None:
            title = f"Anomaly results ({self.model.__class__.__name__})"

        if names_of_scorers is None:
            names_of_scorers = [s.__str__() for s in self.scorers]

        list_window = [s.window for s in self.scorers]

        return show_anomalies_from_scores(
            series=series,
            anomalies=anomalies,
            pred_series=pred_series,
            pred_scores=pred_scores,
            window=list_window,
            names_of_scorers=names_of_scorers,
            title=title,
            metric=metric,
            component_wise=component_wise,
        )

    @property
    def scorers_are_univariate(self):
        """Whether any of the Scorers is univariate."""
        return any(s.is_univariate for s in self.scorers)

    @property
    def scorers_are_trainable(self):
        """Whether any of the Scorers is trainable."""
        return any(s.is_trainable for s in self.scorers)

    @abstractmethod
    def _fit_core(
        self,
        series: Sequence[TimeSeries],
        allow_model_training: bool,
        **kwargs,
    ):
        """Abstract method to implement the model and scorer training."""
        pass

    def _fit_scorers(
        self, list_series: Sequence[TimeSeries], list_pred: Sequence[TimeSeries]
    ):
        """Train the fittable scorers using model forecasts"""
        for scorer in self.scorers:
            if scorer.is_trainable:
                scorer.fit_from_prediction(list_series, list_pred)

    @staticmethod
    def _process_input_series(
        series: Union[TimeSeries, Sequence[TimeSeries]], **kwargs
    ):
        """Checks input series and coverts series and covariates in `kwargs` to sequences."""
        series = _check_input(series, name="series")
        for cov_name in ["past_covariates", "future_covariates"]:
            cov = kwargs.pop(cov_name, None)
            if cov is not None:
                cov = _check_input(cov, name=cov_name)
                _assert_same_length(series, cov, "series", cov_name)
                kwargs[cov_name] = cov
        return series, kwargs



"""
Anomaly Detectors
-----------------

Detectors provide binary anomaly classification on time series. They can typically be used to transform anomaly scores
time series into binary anomaly time series.

Some detectors are trainable. For instance, :class:`~darts.ad.detectors.quantile_detector.QuantileDetector` emits a
binary anomaly for every time step where the observed value(s) are beyond the quantile(s) observed on the training
series.

The main functions are `fit()` (for the trainable detectors), `detect()` and `eval_metric()`.

`fit()` trains the detector over the history of one or multiple time series. It can for instance be called on series
containing anomaly scores (or even raw values) during normal times. The function `detect()` takes an anomaly score
time series as input, and applies the detector to obtain binary predictions. The function `eval_metric()` returns
the accuracy metric ("accuracy", "precision", "recall" or "f1") between a binary prediction time series and some known
binary ground truth time series indicating the presence of anomalies.
"""

from darts.ad.detectors.iqr_detector import IQRDetector
from darts.ad.detectors.quantile_detector import QuantileDetector
from darts.ad.detectors.threshold_detector import ThresholdDetector

__all__ = [
    "IQRDetector",
    "QuantileDetector",
    "ThresholdDetector",
]



from collections.abc import Mapping, Sequence
from typing import Any, Union

import numpy as np

from darts import TimeSeries
from darts.dataprocessing.transformers import BaseDataTransformer
from darts.utils.timeseries_generation import constant_timeseries


class TestBaseDataTransformer:
    class DataTransformerMock(BaseDataTransformer):
        def __init__(
            self,
            scale: float,
            translation: float,
            stack_samples: bool = False,
            mask_components: bool = True,
            parallel_params: Union[bool, Sequence[str]] = False,
        ):
            """
            Applies the transform `transformed_series = scale * series + translation`.

            Parameters
            ----------
            scale
                Scale coefficient of transform.
            translation
                Translational constant of transform.
            stack_samples
                Whether to call `stack_samples` inside of `ts_transform`.
            mask_components
                Whether to automatically apply any provided `component_mask` key word arguments. See
                `BaseDataTransformer` docstring for further details.
            parallel_params
                Specifies which parameters should vary between different parallel jobs, supposing that
                multiple time series are given to `ts_transform`. See `BaseDataTransformer` docstring
                for further details.

            """
            self._scale = scale
            self._translation = translation
            self._stack_samples = stack_samples
            self._mask_components = mask_components
            super().__init__(
                name="DataTransformerMock",
                mask_components=mask_components,
                parallel_params=parallel_params,
            )

        @staticmethod
        def ts_transform(
            series: TimeSeries, params: Mapping[str, Any], **kwargs
        ) -> TimeSeries:
            """
            Implements the transform `scale * series + translation`.

            If `component_mask` is in `kwargs`, this is manually applied and unapplied. If
            `_stack_samples = True` in `params['fixed']`, then `stack_samples` and `unstack_samples`
            all used when computing this transformation.

            """
            fixed_params = params["fixed"]
            stack_samples = fixed_params["_stack_samples"]
            mask_components = fixed_params["_mask_components"]
            scale, translation = fixed_params["_scale"], fixed_params["_translation"]

            # Ensure `component_mask` not passed via `kwargs` if specified `mask_components=True`
            # when transform created
            if mask_components:
                assert "component_mask" not in kwargs

            # Ensure manual masking only performed when `mask_components = False`
            # when transform constructed:
            if not mask_components and ("component_mask" in kwargs):
                vals = TestBaseDataTransformer.DataTransformerMock.apply_component_mask(
                    series, kwargs["component_mask"], return_ts=False
                )
            else:
                vals = series.all_values()

            if stack_samples:
                vals = TestBaseDataTransformer.DataTransformerMock.stack_samples(vals)
            vals = scale * vals + translation
            if stack_samples:
                vals = TestBaseDataTransformer.DataTransformerMock.unstack_samples(
                    vals, series=series
                )

            if not mask_components and ("component_mask" in kwargs):
                vals = (
                    TestBaseDataTransformer.DataTransformerMock.unapply_component_mask(
                        series, vals, kwargs["component_mask"]
                    )
                )

            return series.with_values(vals)

    def test_input_transformed_single_series(self):
        """
        Tests for correct transformation of single series.
        """
        test_input = constant_timeseries(value=1, length=10)

        mock = self.DataTransformerMock(scale=2, translation=10)

        transformed = mock.transform(test_input)

        # 2 * 1 + 10 = 12
        expected = constant_timeseries(value=12, length=10)
        assert transformed == expected

    def test_input_transformed_multiple_series(self):
        """
        Tests for correct transformation of multiple series when
        different param values are used for different parallel
        jobs (i.e. test that `parallel_params` argument is treated
        correctly). Also tests that transformer correctly handles
        being provided with fewer input series than fixed parameter
        value sets.
        """
        test_input_1 = constant_timeseries(value=1, length=10)
        test_input_2 = constant_timeseries(value=2, length=11)

        # Don't have different params for different jobs:
        mock = self.DataTransformerMock(scale=2, translation=10, parallel_params=False)
        (transformed_1, transformed_2) = mock.transform((test_input_1, test_input_2))
        # 2 * 1 + 10 = 12
        assert transformed_1 == constant_timeseries(value=12, length=10)
        # 2 * 2 + 10 = 14
        assert transformed_2 == constant_timeseries(value=14, length=11)

        # Have different `scale` param for different jobs:
        mock = self.DataTransformerMock(
            scale=(2, 3), translation=10, parallel_params=["_scale"]
        )
        (transformed_1, transformed_2) = mock.transform((test_input_1, test_input_2))
        # 2 * 1 + 10 = 12
        assert transformed_1 == constant_timeseries(value=12, length=10)
        # 3 * 2 + 10 = 16
        assert transformed_2 == constant_timeseries(value=16, length=11)

        # If only one timeseries provided, should apply parameters defined for
        # for the first to that series:
        transformed_1 = mock.transform(test_input_1)
        assert transformed_1 == constant_timeseries(value=12, length=10)

        # Have different `scale`, `translation`, and `stack_samples` params for different jobs:
        mock = self.DataTransformerMock(
            scale=(2, 3),
            translation=(10, 11),
            stack_samples=(False, True),
            mask_components=(False, False),
            parallel_params=True,
        )
        (transformed_1, transformed_2) = mock.transform((test_input_1, test_input_2))
        # 2 * 1 + 10 = 12
        assert transformed_1 == constant_timeseries(value=12, length=10)
        # 3 * 2 + 11 = 17
        assert transformed_2 == constant_timeseries(value=17, length=11)

        # If only one timeseries provided, should apply parameters defined for
        # for the first to that series:
        transformed_1 = mock.transform(test_input_1)
        assert transformed_1 == constant_timeseries(value=12, length=10)

        # Specify three sets of fixed params, but pass only one or two series as inputs
        # to `transform`; transformer should apply `i`th set of fixed params to the `i`th
        # input passed to `transform`
        mock = self.DataTransformerMock(
            scale=(2, 3, 4),
            translation=(10, 11, 12),
            stack_samples=(False, True, False),
            mask_components=(False, False, False),
            parallel_params=True,
        )
        # If single series provided to transformer with three sets of
        # fixed params, should transform using the first set of fixed
        # parameters:
        transformed_1 = mock.transform(test_input_1)
        assert transformed_1 == constant_timeseries(value=12, length=10)
        # If two series provided to transformer with three sets of
        # fixed params, should transform using the first and second set of fixed
        # parameters:
        transformed_1, transformed_2 = mock.transform((test_input_1, test_input_2))
        assert transformed_1 == constant_timeseries(value=12, length=10)
        assert transformed_2 == constant_timeseries(value=17, length=11)

    def test_input_transformed_multiple_samples(self):
        """
        Tests that `stack_samples` and `unstack_samples` correctly
        implemented when considering multi-sample timeseries.
        """
        test_input = constant_timeseries(value=1, length=10)
        test_input = test_input.concatenate(
            constant_timeseries(value=2, length=10), axis="sample"
        )

        mock = self.DataTransformerMock(scale=2, translation=10, stack_samples=True)
        transformed = mock.transform(test_input)

        # 2 * 1 + 10 = 12
        expected = constant_timeseries(value=12, length=10)
        # 2 * 2 + 10 = 14
        expected = expected.concatenate(
            constant_timeseries(value=14, length=10), axis="sample"
        )
        assert transformed == expected

    def test_input_transformed_masking(self):
        """
        Tests that automatic component masking is correctly implemented,
        and that manual component masking is also handled correctly
        through `kwargs` + `apply_component_mask`/`unapply_component_mask`
        methods.
        """
        test_input = TimeSeries.from_values(np.ones((4, 3, 5)))
        mask = np.array([True, False, True])
        # Second component should be untransformed:
        scale = 2
        translation = 10
        expected = np.stack(
            [12 * np.ones((4, 5)), np.ones((4, 5)), 12 * np.ones((4, 5))], axis=1
        )
        expected = TimeSeries.from_values(expected)

        # Automatically apply component mask:
        mock = self.DataTransformerMock(
            scale=scale, translation=translation, mask_components=True
        )
        transformed = mock.transform(test_input, component_mask=mask)
        assert transformed == expected

        # Manually apply component mask:
        mock = self.DataTransformerMock(scale=2, translation=10, mask_components=False)
        transformed = mock.transform(test_input, component_mask=mask)
        assert transformed == expected
